from re import L
import matplotlib as mpl


# temp
import bluebelt
import plotly.graph_objects as go


def linestyle(value):

    # matplotlib "solid", "dotted", "dashed", "dashdot"
    # plotly "solid", "dot", "dash", "dashdot"

    values = {"solid": "solid", "dotted": "dot", "dashed": "dash", "dashdot": "dashdot"}

    if value in values.keys():
        return values.get(value)
    else:
        return None

def marker(value):
    values = {
        "o": "circle",
        "v": "triangle-down",
        "^": "triangle-up",
        "<": "triangle-left",
        ">": "triangle-right",
        "+": "cross",
        "1": "y-down",
        "2": "y-up",
        "3": "y-left",
        "4": "y-right",
        "s": "square",
        "p": "pentagon",
        "8": "octagon",
        "*": "star",
        "x": "x",
        "D": "diamond",
        "d": "diamond-tall",
        }

    return values.get(value, "circle")
    
def tick_direction(value):
    return {"in": "inside", "out": "outside", "inout": "outside"}.get(value)


def get_margin(title=None):
    return dict(
        l=0,
        r=0,
        t=30 if title else 0,
        b=0,
    )


def xaxis(xlabel=None, xlim=None, format_xticks=None):
    return dict(
        title=xlabel,
        gridcolor=mpl.colors.to_hex(mpl.rcParams["grid.color"])
        if mpl.rcParams["axes.grid.axis"] in ["x", "both"]
        else None,
        gridwidth=mpl.rcParams["grid.linewidth"]
        if mpl.rcParams["axes.grid.axis"] in ["x", "both"]
        else None,
        showgrid=True
        if mpl.rcParams["axes.grid.which"] in ["major", "both"]
        else False,
        ticks=tick_direction(mpl.rcParams["xtick.direction"])
        if mpl.rcParams["axes.spines.bottom"]
        else "",
        tickwidth=mpl.rcParams["xtick.major.width"],
        tickcolor=mpl.colors.to_hex(mpl.rcParams["xtick.color"]),
        ticklen=mpl.rcParams["xtick.major.size"],
        tickformat=format_xticks,
        showline=mpl.rcParams["axes.spines.bottom"],
        linecolor=mpl.colors.to_hex(mpl.rcParams["axes.edgecolor"]),
        linewidth=mpl.rcParams["axes.linewidth"],
        mirror=mpl.rcParams["xtick.major.top"],
        range=xlim,
    )


def yaxis(ylabel=None, ylim=None, format_yticks=None):
    return dict(
        title=ylabel,
        gridcolor=mpl.colors.to_hex(mpl.rcParams["grid.color"])
        if mpl.rcParams["axes.grid.axis"] in ["y", "both"]
        else None,
        gridwidth=mpl.rcParams["grid.linewidth"]
        if mpl.rcParams["axes.grid.axis"] in ["y", "both"]
        else None,
        showgrid=True
        if mpl.rcParams["axes.grid.which"] in ["major", "both"]
        else False,
        ticks=tick_direction(mpl.rcParams["ytick.direction"])
        if mpl.rcParams["axes.spines.left"]
        else "",
        tickwidth=mpl.rcParams["ytick.major.width"],
        tickcolor=mpl.colors.to_hex(mpl.rcParams["ytick.color"]),
        ticklen=mpl.rcParams["ytick.major.size"],
        tickformat=format_yticks,
        showline=mpl.rcParams["axes.spines.left"],
        linecolor=mpl.colors.to_hex(mpl.rcParams["axes.edgecolor"]),
        linewidth=mpl.rcParams["axes.linewidth"],
        mirror=mpl.rcParams["ytick.major.right"],
        range=ylim,
    )


def layout(
    title=None,
    xlabel=None,
    ylabel=None,
    xlim=None,
    ylim=None,
    format_xticks=None,
    format_yticks=None,
    legend=True,
    **kwargs,
):

    layout = dict(
        title=title,
        plot_bgcolor=mpl.rcParams["figure.facecolor"],
        hovermode="closest",
        margin=get_margin(title=title),
        xaxis=xaxis(xlabel=xlabel, xlim=xlim, format_xticks=format_xticks),
        yaxis=yaxis(ylabel=ylabel, ylim=ylim, format_yticks=format_yticks),
        legend=dict(yanchor="top", y=0.99, xanchor="right", x=0.99),
        showlegend=legend,
    )
    if not kwargs.get('dash', False):
        layout['width']=mpl.rcParams["figure.figsize"][0] * mpl.rcParams["figure.dpi"]
        layout['height']=mpl.rcParams["figure.figsize"][1] * mpl.rcParams["figure.dpi"]
    else:
        layout['paper_bgcolor']='rgba(0,0,0,0)'
        layout['plot_bgcolor']='rgba(0,0,0,0)'
        layout['legend'] = dict(bgcolor='rgba(0,0,0,0)', yanchor="top", xanchor="right")

    return layout

def to_rgba(hex, alpha):
    return 'rgba' + str(mpl.colors.to_rgba(hex +  format(int(255*alpha), 'x')))


def _default_plotly():
    if bluebelt.config('plotting') == 'matplotlib':
        pass
    elif bluebelt.config('plotting') == 'plotly':

        layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    format_xticks=format_xticks,
                    legend=legend,
                    **kwargs,
                )
            )

        line_width = np.maximum(1, bluebelt.style("equal_means.boxplot.boxes.linewidth"))
        
        data = [
            go.Box(
                x=plot_obj.frame[col],
                y0=idx,
                marker=dict(color=bluebelt.style("equal_means.boxplot.fliers.color"), symbol=m2p.marker(bluebelt.style("equal_means.boxplot.fliers.marker"))),
                fillcolor=bluebelt.style("equal_means.boxplot.boxes.facecolor"),
                line_color=bluebelt.style("equal_means.boxplot.boxes.edgecolor"),
                line_width=line_width,
                hoverinfo='skip',
                showlegend=False,
                name=col,
            ) for idx, col in enumerate(plot_obj.frame.columns[::-1])
        ]

        fig = go.Figure(data=data, layout=layout)

        return fig
    
    else:
        return


import pandas as pd

def is_series(_obj):
    if not isinstance(_obj, pd.Series):
        raise ValueError(f'The object is expected to be a pandas Series, not {type(_obj).__name__}')
    
def is_frame(_obj):
    if not isinstance(_obj, pd.DataFrame):
        raise ValueError(f'The object is expected to be a pandas DataFrame, not {type(_obj).__name__}')
    
def is_series_or_frame(_obj):
    if not isinstance(_obj, (pd.Series, pd.DataFrame)):
        raise ValueError(f'The object is expected to be a pandas Series or DataFrame, not {type(_obj).__name__}')
    
def is_index(_obj):
    if not isinstance(_obj, pd.Index):
        raise ValueError(f'The object is expected to be a pandas Index, not {type(_obj).__name__}')
    
def is_datetimeindex(_obj):
    if not isinstance(_obj, pd.DatetimeIndex):
        raise ValueError(f'The object is expected to be a pandas DatetimeIndex, not {type(_obj).__name__}')

def is_multiindex(_obj):
    if not isinstance(_obj, pd.DatetimeIndex):
        raise ValueError(f'The object is expected to be a pandas MultiIndex, not {type(_obj).__name__}')

def has_datetimeindex(_obj):
    if not isinstance(_obj.index, pd.DatetimeIndex):
        raise ValueError(f'{type(_obj).__name__} is expected to have a DateTimeIndex, not a {type(_obj.index).__name__}')

def has_multiindex(_obj):
    if not isinstance(_obj.index, pd.DatetimeIndex):
        raise ValueError(f'{type(_obj).__name__} is expected to have a MultiIndex, not a {type(_obj.index).__name__}')


import pandas as pd
import math
from matplotlib.colors import to_rgb, to_hex, is_color_like

def _get_colors_from_style(style_dict):
    result = []
    for value in style_dict.values():
        if isinstance(value, dict):
            result += _get_colors_from_style(value)
        elif is_color_like(value):
            return [value]
    return list(set(result))

def _closest_colors(match_colors, to_colors):
    
    if len(to_colors) < len(match_colors):
        raise ValueError(f"Not enough colors to match the existing colors. The list has {len(to_colors)} colors but needs at least {len(match_colors)} colors.")
        
    if not isinstance(match_colors, list):
        match_colors = [match_colors]
    
    frame = pd.DataFrame([to_rgb(c) for c in to_colors])
    result = {}
        
    def close(row, color):
        r, g, b = to_rgb(color)
        return math.sqrt((r - row.iloc[0])**2 + (g - row.iloc[1])**2 + (b - row.iloc[2])**2)
    
    to_colors = pd.DataFrame(to_colors)
    
    for c in match_colors:
                
        # get closest color and drop from frame
        idx = frame.apply(lambda x: close(x, c), axis=1).idxmin()
        closest = to_hex(tuple(frame.iloc[idx].values))
        frame = frame.drop(index = idx).reset_index(drop=True)

        # add to result if the color is not the same
        if c != closest:
            result[c] = closest
        
    return result

def _replace_colors(style, colors_dict):
    replaced = {}
    for key, value in style.items():
        if isinstance(value, dict):
            replaced[key] = _replace_colors(value, colors_dict)
        elif isinstance(value, list):
            replaced[key] = [colors_dict.get(val, val) if is_color_like(val) else val for val in value]
        elif is_color_like(value):
            # replace the color if it is in the colors_dict
            replaced[key] = colors_dict.get(value, value)
        else:
            replaced[key] = value
    return replaced

def replace_colors(style, to_colors):
    match_colors = style.get('colors', _get_colors_from_style(style))
    colors_dict = _closest_colors(match_colors, to_colors)
    return _replace_colors(style, colors_dict)



import numpy as np
import pandas as pd

import bluebelt.helpers.check as check
import bluebelt.core.series

import bluebelt.analysis.forecast

import bluebelt.statistics.hypothesis_testing
import bluebelt.statistics.basic


@pd.api.extensions.register_dataframe_accessor("_")
@pd.api.extensions.register_dataframe_accessor("blue")
class DataFrameToolkit:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.stats = self.stats(self._obj)        

        # copy Series functions
        self.weekend = bluebelt.core.series.SeriesToolkit(self._obj).weekend
        self.not_weekend = bluebelt.core.series.SeriesToolkit(self._obj).not_weekend
        self.weekdays = bluebelt.core.series.SeriesToolkit(self._obj).weekdays
        self.holiday = bluebelt.core.series.SeriesToolkit(self._obj).holiday
        self.holidays = bluebelt.core.series.SeriesToolkit(self._obj).holidays
        self.not_holiday = bluebelt.core.series.SeriesToolkit(self._obj).not_holiday
        self.not_holidays = bluebelt.core.series.SeriesToolkit(self._obj).not_holidays
        self.workdays = bluebelt.core.series.SeriesToolkit(self._obj).workdays

        self.resample = bluebelt.core.series.SeriesToolkit(self._obj).resample
        self.flatten = bluebelt.core.series.SeriesToolkit(self._obj).flatten
        self.project = bluebelt.core.series.SeriesToolkit(self._obj).project

        self.line = bluebelt.core.series.SeriesToolkit(self._obj).line
        self.bar = bluebelt.core.series.SeriesToolkit(self._obj).bar
        self.area = bluebelt.core.series.SeriesToolkit(self._obj).area
        self.hist = bluebelt.core.series.SeriesToolkit(self._obj).hist
        self.dist = bluebelt.core.series.SeriesToolkit(self._obj).dist
        self.histogram = bluebelt.core.series.SeriesToolkit(self._obj).hist
        self.box = bluebelt.core.series.SeriesToolkit(self._obj).box
        self.boxplot = bluebelt.core.series.SeriesToolkit(self._obj).box

        # copy Series classes
        self.planning = bluebelt.core.series.SeriesToolkit(self._obj).planning
        self.forecast = bluebelt.core.series.SeriesToolkit(self._obj).forecast

    def fill_na(self):
        """
        Fill all missing values of a pandas DataFrame with the
        interpolated values of the previous and next weeks weekday.

        Parameters
        ----------
        self: DataFrame

        Returns
        -------
        pandas.DataFrame
        """
        result = pd.DataFrame(index=self._obj.index)

        for column in self._obj:
            result[column] = bluebelt.core.series.SeriesToolkit(
                self._obj[column]
            ).fill_na()

        return result

    def project(self, year=None, adjust_holidays=True):
        return bluebelt.data.projection.project(
            self._obj, year=year, adjust_holidays=adjust_holidays
        )

    def subset(self, inverse=False, **kwargs):
        """
        Create a subset based on the pandas.Dataframe column values.

        Parameters
        ----------
        frame: pandas.DataFrame
        inverse: bool, default False
            inverse the result
        kwargs: str=*
            columns names as arguments and values as values or a list of values
            any space in the column name can be replaced with an underscore (_)

        Returns
        -------
        Subset


        Example
        -------

        df
            | column_1  | column_2  | column_3  |
        -----------------------------------------
        0   | a         | a         | a         |
        1   | b         | a         | b         |
        2   | a         | b         | c         |
        3   | b         | b         | a         |
        4   | a         | c         | b         |
        5   | b         | c         | c         |
        6   | a         | a         | a         |
        7   | b         | a         | b         |
        8   | a         | b         | c         |
        9   | b         | b         | a         |

        >> subset = bluebelt.data.subsets.Subset(frame, column_1="a", column_2=["a", "b"])
        >> subset

            | column_1  | column_2  | column_3  |
        -----------------------------------------
        0   | a         | a         | a         |
        2   | a         | b         | c         |
        6   | a         | a         | a         |
        8   | a         | b         | c         |
        """

        # build filters
        filters = {}
        for col in kwargs:
            if col in self._obj.columns:
                values = (
                    kwargs.get(col)
                    if isinstance(kwargs.get(col), list)
                    else [kwargs.get(col)]
                )
                for value in values:
                    if value not in self._obj[col].values:
                        raise ValueError(f"{value} is not in {col}")
                filters[col] = values
            elif col.replace("_", " ") in self._obj.columns:
                _col = col.replace("_", " ")
                values = (
                    kwargs.get(col)
                    if isinstance(kwargs.get(col), list)
                    else [kwargs.get(col)]
                )
                for value in values:
                    if value not in self._obj[_col].values:
                        raise ValueError(f"{value} is not in {_col}")
                filters[_col] = values
            else:
                raise ValueError(f"{col} is not in frame")

        self.filters = filters

        # filter the frame
        if inverse:
            frame = self._obj[
                self._obj.isin(filters).sum(axis=1) != len(filters.keys())
            ]
        else:
            frame = self._obj[
                self._obj.isin(filters).sum(axis=1) == len(filters.keys())
            ]

        return frame
    
    class stats:
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        def dagostino_pearson(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(
                self._obj, alpha=alpha
            )

        normal_distribution = dagostino_pearson

        def anderson_darling(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(
                self._obj, alpha=alpha
            )

        def std_within(self, how=None, observations=2, **kwargs):
            return bluebelt.statistics.std.StdWithin(
                self._obj, how=how, observations=observations, **kwargs
            )

        def one_sample_t(self, popmean=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.OneSampleT(
                self._obj, popmean=popmean, alpha=alpha, **kwargs
            )

        def wilcoxon(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(
                self._obj, alpha=alpha, **kwargs
            )

        def equal_means(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualMeans(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def two_sample_t(
            self, columns=None, related=False, confidence=0.95, alpha=0.05, **kwargs
        ):
            return bluebelt.statistics.hypothesis_testing.TwoSampleT(
                self._obj,
                columns=columns,
                related=related,
                confidence=confidence,
                alpha=alpha,
                **kwargs,
            )

        def mann_whitney(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.MannWhitney(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def anova(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Anova(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def kruskal(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.KruskalWallis(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        kruskal_wallis = kruskal

        def levene(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Levene(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def bartlett(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Bartlett(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def f_test(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.FTest(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        def equal_variances(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualVariances(
                self._obj, columns=columns, alpha=alpha, **kwargs
            )

        equal_var = equal_variances

        def correlation(self, columns=None, confidence=0.95, **kwargs):
            return bluebelt.statistics.basic.Correlation(
                self._obj, columns=columns, confidence=confidence
            )



import datetime
import numpy as np

import matplotlib.ticker
import matplotlib.dates

def ticks(_obj, ax, **kwargs):

    max_xticks = kwargs.get('max_xticks', get_max_xticks(ax))
    #ax.minorticks_off()

    # set year ticks at major positions
    ticks = _obj.index.unique()
    step = int(np.ceil(len(ticks)/np.minimum(len(ticks), max_xticks)))
    ticklabels = ticks[::step]
    locator_major = matplotlib.ticker.FixedLocator(range(1, len(ticks)+1, step))
    ax.xaxis.set_major_locator(locator_major)
    ax.set_xticklabels(ticklabels)
    
def week(_obj, ax, **kwargs):

    max_xticks = kwargs.get('max_xticks', get_max_xticks(ax))
        
    # set year ticks at major positions
    year_ticks = []#get_year_ticks(_obj, max_xticks=max_xticks)
    locator_major = matplotlib.ticker.FixedLocator(year_ticks)
    formatter_major = matplotlib.dates.DateFormatter('%V\n%G')
    ax.xaxis.set_major_locator(locator_major)
    ax.xaxis.set_major_formatter(formatter_major)

    # set week ticks at minor positions
    week_ticks = get_week_ticks(_obj, max_xticks=max_xticks)
    locator_minor = matplotlib.ticker.FixedLocator(week_ticks)
    formatter_minor = matplotlib.dates.DateFormatter('%V')
    ax.xaxis.set_minor_locator(locator_minor)
    ax.xaxis.set_minor_formatter(formatter_minor)

def year_week(_obj, ax, **kwargs):

    max_xticks = kwargs.get('max_xticks', get_max_xticks(ax))

    # set year ticks at major positions
    year_ticks = get_year_ticks(_obj, max_xticks=max_xticks)
    locator_major = matplotlib.ticker.FixedLocator(year_ticks)
    formatter_major = matplotlib.dates.DateFormatter('%V\n%G')
    ax.xaxis.set_major_locator(locator_major)
    ax.xaxis.set_major_formatter(formatter_major)

    # set week ticks at minor positions
    week_ticks = get_week_ticks(_obj, max_xticks=max_xticks)
    locator_minor = matplotlib.ticker.FixedLocator(week_ticks)
    formatter_minor = matplotlib.dates.DateFormatter('%V')
    ax.xaxis.set_minor_locator(locator_minor)
    ax.xaxis.set_minor_formatter(formatter_minor)


def get_year_ticks(obj, max_xticks=17):
    
    if max_xticks is not None:
        rng = obj.index.isocalendar().year.unique().size
        _group = np.ceil(rng/max_xticks)
        _groups = np.array([1, 4, 13, 26, 52])
        group = _groups[np.greater_equal(_groups - _group, 0).argmax()]
    else:
        group = 1
        
    ticks = []
    years = obj.index.isocalendar().year.unique()
    if group > 1:
        years = years[np.where(np.mod(years, group)==1)]
        
    for year in years:#obj.index.isocalendar().year.unique():
        ticks.append(matplotlib.dates.date2num(datetime.datetime.fromisocalendar(year, 1, 1)))
            
    return ticks

def get_week_ticks(obj, max_xticks=17):
    
    if max_xticks is not None:
        rng = obj.index.isocalendar()[['year', 'week']].apply(tuple, axis=1).unique().size
        _group = np.ceil(rng/max_xticks)
        _groups = np.array([1, 4, 13, 26, 52])
        group = _groups[np.greater_equal(_groups - _group, 0).argmax()]
    else:
        group = 1
    
    ticks = []
    for year in obj.index.isocalendar().year.unique():
        weeks = obj[obj.index.isocalendar().year==year].index.isocalendar().week.unique().to_numpy()
        if group > 1:
            weeks = weeks[np.where(np.mod(weeks, group)==1)]
        for week in weeks:
            if week < 53: # that would mess things up
                ticks.append(matplotlib.dates.date2num(datetime.datetime.fromisocalendar(year, week, 1)))
            
    return ticks

def get_max_xticks(ax, ratio=0.75):
    
    # get ax width in pixels
    ax_width = ax.figure.get_figwidth() * (ax.figure.subplotpars.right - ax.figure.subplotpars.left) * ax.figure.dpi
    
    # get text width: tick label font size * 2 characters * .75 width ratio
    label_width = ax.xaxis.get_majorticklabels()[0].get_fontsize() * 2 * ratio
    
    return np.floor(ax_width / label_width)


import os
import pandas as pd
import numpy as np

import bluebelt.helpers.ci
import bluebelt.helpers.boxplot

import scipy.stats as stats

import matplotlib as mpl
import matplotlib.pyplot as plt
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import bluebelt.helpers.matplotlib2plotly as m2p

import math


def index():
    df = pd.DataFrame(
        [
            ["discrete", "discrete", "-", "-", "-", "chi square test", ""],
            ["discrete", "continuous", "-", "-", "-", "logistic regression", ""],
            [
                "continuous",
                "discrete",
                "mean",
                "1",
                "normal",
                "1 sample t-test",
                "one_sample_t",
            ],
            [
                "continuous",
                "discrete",
                "mean",
                "1",
                "non-normal",
                "1 sample Wilcoxon test",
                "wilcoxon",
            ],
            [
                "continuous",
                "discrete",
                "mean",
                "2",
                "normal",
                "2 sample t-test",
                "two_sample_t",
            ],
            [
                "continuous",
                "discrete",
                "mean",
                "2",
                "non-normal",
                "Mann-Whitney test",
                "mann_whitney",
            ],
            ["continuous", "discrete", "mean", ">2", "normal", "1 way Anova", "anova"],
            [
                "continuous",
                "discrete",
                "mean",
                ">2",
                "non-normal",
                "Kruskal-Wallis test",
                "kruskal",
            ],
            ["continuous", "discrete", "variance", "2", "normal", "F-test", "f_test"],
            [
                "continuous",
                "discrete",
                "variance",
                "2",
                "non-normal",
                "Levene’s test",
                "levene",
            ],
            [
                "continuous",
                "discrete",
                "variance",
                ">2",
                "normal",
                "Bartlett’s test",
                "bartlett",
            ],
            [
                "continuous",
                "discrete",
                "variance",
                ">2",
                "non-normal",
                "Levene’s test",
                "levene",
            ],
            ["continuous", "continuous", "", "-", "-", "regression", ""],
        ],
        columns=[
            "Y",
            "X",
            "investigate",
            "# data groups",
            "distribution",
            "test",
            "bluebelt",
        ],
    )
    print(df.to_string(index=False))


class OneSampleT:
    def __init__(
        self, frame, columns=None, popmean=None, confidence=0.95, alpha=0.05, **kwargs
    ):

        if not popmean:
            raise ValueError(
                "Please provide a value for the population mean. e.g. popmean=100"
            )

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame).dropna()
        elif isinstance(columns, str):
            frame = frame[columns].dropna()
        elif isinstance(columns, list):
            frame = frame[columns[:1]].dropna()
        else:
            frame = frame.iloc[:, :1].dropna()

        self.frame = frame
        self.nrows = frame.shape[0]
        self.ncols = 1
        self.popmean = popmean
        self.confidence = confidence
        self.alpha = alpha
        self.test = "One Sample t-Test"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.ttest_1samp(
            a=self.frame.iloc[:, 0].values, popmean=self.popmean, **kwargs
        )
        self.ci_mean = bluebelt.helpers.ci.ci_mean(self.frame.iloc[:, 0])
        self.passed = True if self.p_value > self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'
        _result += f'  {"population mean:":<30}{self.popmean}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, confidence={self.confidence}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def result(self):
        print(self.statistic)

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class Wilcoxon:
    """
    Calculate the Wilcoxon signed-rank test.

    The Wilcoxon signed-rank test tests the null hypothesis that two related paired samples come from the same distribution.
    In particular, it tests whether the distribution of the differences x - y is symmetric about zero.
    It is a non-parametric version of the paired T-test.
    This test is performed with the assumption that if a pandas Series is provided the series contains the differences between x and y.
    """

    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)
        elif isinstance(columns, str):
            frame = frame[columns]
        elif isinstance(columns, list):
            frame = frame[columns[:2]]
        else:
            frame = frame.iloc[:, :2]

        self.frame = frame
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Wilcoxon Signed-Rank Test"
        self.h0 = f"$H_0:$" + "difference has symmetric distribution around zero"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.array_a = self.frame.iloc[:, 0].dropna()
        self.array_b = self.frame.iloc[:, 1].dropna() if self.ncols == 2 else None
        self.statistic, self.p_value = stats.wilcoxon(
            x=self.array_a, y=self.array_b, **kwargs
        )
        self.passed = True if self.p_value > self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, confidence={self.confidence}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class TwoSampleT:
    def __init__(
        self, frame, columns=None, related=False, confidence=0.95, alpha=0.05, **kwargs
    ):

        if isinstance(columns, list):
            frame = frame[columns[:2]]
        else:
            frame = frame.iloc[:, :2]

        self.frame = frame
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.related = related
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Two Sample t-Test"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.array_a = self.frame.iloc[:, 0].dropna()
        self.array_b = self.frame.iloc[:, 1].dropna()

        self.mean_a = self.array_a.mean()
        self.mean_b = self.array_b.mean()

        self.max_a = self.array_a.max()
        self.max_b = self.array_b.max()

        self.min_a = self.array_a.min()
        self.min_b = self.array_b.min()

        self.ci_mean_a = bluebelt.helpers.ci.ci_mean(self.array_a)
        self.ci_mean_b = bluebelt.helpers.ci.ci_mean(self.array_b)

        # related or not
        if self.related:
            statistic, pvalue = stats.ttest_rel(self.array_a, self.array_b, **kwargs)
        else:
            # test equal variance
            equal_var = (
                True if Levene(frame=self.frame, alpha=self.alpha).passed else False
            )
            statistic, pvalue = stats.ttest_ind(
                self.array_a, self.array_b, equal_var=equal_var, **kwargs
            )

        self.statistic = statistic
        self.p_value = pvalue
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, confidence={self.confidence}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def result(self):
        print(self.statistic)

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class MannWhitney:
    """
    apply the Mann-Whitney test on a pandas dataframe with the columns as groups to compare
    Returns the scipy Mann-Whitney; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """

    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        if isinstance(columns, list):
            frame = frame[columns[:2]]
        else:
            frame = frame.iloc[:, :2]

        self.frame = frame
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Mann-Whitney"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        self.array_a = self.frame.iloc[:, 0].dropna().values
        self.array_b = self.frame.iloc[:, 1].dropna().values

        self.statistic, self.p_value = stats.mannwhitneyu(self.array_a, self.array_b)
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, confidence={self.confidence}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class Anova:
    """
    apply the one way ANOVA test on a pandas dataframe with the columns as groups to compare
    Returns the scipy F_onewayResult; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """

    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "ANOVA"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        self.statistic, self.p_value = stats.f_oneway(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class KruskalWallis:
    """
    apply the Kruskal-Wallis test on a pandas dataframe with the columns as groups to compare
    Returns the scipy KruskalResult; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """

    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Kruskal-Wallis"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        self.statistic, self.p_value = stats.kruskal(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, confidence={self.confidence}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


class EqualMeans:
    """
    Test a dataframe for equal means over the dataframe columns.

    frame: a pandas DataFrame
    columns: string or list of strings with the column names to be tested
    alpha: test threshold

    applies ANOVA or Kruskal-Wallis test

    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means

    """

    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        # test for normal distribution; null hypothesis: values come from a normal distribution
        if (
            NormalDistribution(frame=self.frame, alpha=self.alpha).passed
            and Levene(frame=self.frame, alpha=self.alpha).passed
        ):
            # all columns in the dataframe come from a normal distribution AND have equal variances
            # do Anova
            test = Anova(frame=self.frame, alpha=self.alpha)
        else:
            # not all columns in the dataframe come from a normal distribution OR have equal variances
            # do Kruskal-Wallis
            test = KruskalWallis(frame=self.frame, alpha=self.alpha)

        self.test = test.test
        self.statistic = test.statistic
        self.p_value = test.p_value
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"confidence:":<30}{self.confidence}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(
        self,
        xlim=(None, None),
        format_xticks=None,
        format_stats=".2f",
        format_means=".2f",
        title=None,
        xlabel=None,
        ylabel=None,
        legend=True,
        path=None,
        **kwargs,
    ):
        return _equal_means_plot(
            self,
            xlim=xlim,
            format_xticks=format_xticks,
            format_stats=format_stats,
            format_means=format_means,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            legend=legend,
            path=path,
            **kwargs,
        )


def _equal_means_plot(
    plot_obj,
    xlim=(None, None),
    format_xticks=None,
    format_stats=".2f",
    format_means=".2f",
    title=None,
    xlabel=None,
    ylabel=None,
    legend=True,
    path=None,
    **kwargs,
):

    title = title if title is not None else f"{plot_obj.test}"

    if bluebelt.config('plotting') == 'matplotlib':

        # prepare figure
        fig, axes = plt.subplots(
            nrows=plot_obj.ncols,
            ncols=1,
            sharex=True,
            gridspec_kw={"wspace": 0, "hspace": 0},
            **kwargs,
        )

        for row in range(0, plot_obj.ncols):

            if plot_obj.ncols == 1:
                ax = axes
            else:
                ax = axes[row]

            # set right zorder
            ax.set_zorder((plot_obj.ncols - row) * 10)
            # get the data
            array = plot_obj.frame.iloc[:, row].dropna()

            # 1. box plot ############################################
            boxplot = ax.boxplot(array, vert=False, widths=0.3, whis=1.5)
            for box in boxplot["boxes"]:
                box.set(**bluebelt.style("equal_means.boxplot.boxes"))

            # 2. CI for the mean #####################################
            ci_mean = bluebelt.helpers.ci.ci_mean(array, confidence=plot_obj.confidence)

            ax.plot(
                [ci_mean[0], ci_mean[1]],
                [1.5, 1.5],
                **bluebelt.style("equal_means.ci_mean.line"),
            )
            ax.axvline(
                x=ci_mean[0],
                ymin=9 / 14,
                ymax=11 / 14,
                **bluebelt.style("equal_means.ci_mean.vline"),
            )
            ax.axvline(
                x=ci_mean[1],
                ymin=9 / 14,
                ymax=11 / 14,
                **bluebelt.style("equal_means.ci_mean.vline"),
            )
            ax.scatter([array.mean()], [1.5], **bluebelt.style("equal_means.ci_mean.mean"))

            # fill the CI area
            ci_area = np.linspace(ci_mean[0], ci_mean[1], 100)
            ax.fill_between(
                x=(ci_mean[0], ci_mean[1]),
                y1=0,
                y2=2,
                label="CI sample mean",
                **bluebelt.style("equal_means.ci_mean.fill"),
            )
            ax.axvline(
                x=ci_mean[0], ymin=0, ymax=2, **bluebelt.style("equal_means.ci_mean.vline")
            )
            ax.axvline(
                x=ci_mean[1], ymin=0, ymax=2, **bluebelt.style("equal_means.ci_mean.vline")
            )

            # plot CI values
            ax.text(
                ci_mean[0],
                1.5,
                f"{ci_mean[0]:{format_means}} ",
                **bluebelt.style("equal_means.ci_mean.min_text"),
            )
            ax.text(
                ci_mean[1],
                1.5,
                f" {ci_mean[1]:{format_means}}",
                **bluebelt.style("equal_means.ci_mean.max_text"),
            )

            # plot popmean if it exists
            if hasattr(plot_obj, "popmean"):
                ax.axvline(
                    plot_obj.popmean,
                    ymin=0,
                    ymax=2,
                    label="population mean",
                    **bluebelt.style("equal_means.popmean.line"),
                )

            ax.set_ylim([0.25, 2])
            ax.set_yticklabels([array.name])

            # set title and legend
            if row == 0:
                ax.set_title(_get_h0_equal_means(plot_obj, format_stats=format_stats), **bluebelt.style("equal_means.title"))
                # legend
                if legend:
                    ax.legend()
                elif ax.get_legend() is not None:
                    ax.get_legend().set_visible(False)
            elif ax.get_legend() is not None:
                ax.get_legend().set_visible(False)

        # format things
        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[-1]

        # limit axis
        ax.set_xlim(xlim)

        # format ticks
        if format_xticks:
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])

        # labels
        if title:
            fig.suptitle(
                t=title, x=fig.subplotpars.left, **bluebelt.style("equal_means.suptitle")
            )

        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        plt.tight_layout()

        # and again
        if title:
            fig.suptitle(
                t=title, x=fig.subplotpars.left, **bluebelt.style("equal_means.suptitle")
            )

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    elif bluebelt.config('plotting') == 'plotly':

        if kwargs.get('dash', False):
            title = None
        elif title is not None:
            title = _get_h0_equal_means_plotly(plot_obj, title=title, format_stats=format_stats)

        layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    format_xticks=format_xticks,
                    legend=legend,
                    **kwargs,
                )
            )

        line_width = np.maximum(1, bluebelt.style("equal_means.boxplot.boxes.linewidth"))
        
        data = [
            go.Box(
                x=plot_obj.frame[col],
                y0=idx,
                marker=dict(color=bluebelt.style("equal_means.boxplot.fliers.color"), symbol=m2p.marker(bluebelt.style("equal_means.boxplot.fliers.marker"))),
                fillcolor=bluebelt.style("equal_means.boxplot.boxes.facecolor"),
                line_color=bluebelt.style("equal_means.boxplot.boxes.edgecolor"),
                line_width=line_width,
                hoverinfo='skip',
                showlegend=False,
                name=col,
            ) for idx, col in enumerate(plot_obj.frame.columns[::-1])
        ]

        fig = go.Figure(data=data, layout=layout)

        fig.update_layout(boxgap=0.7)

        for idx, col in enumerate(plot_obj.frame.columns[::-1]):
            mean = plot_obj.frame[col].dropna().mean()
            ci_mean = bluebelt.helpers.ci.ci_mean(plot_obj.frame[col].dropna(), confidence=plot_obj.confidence)
            fig.add_trace(
                go.Scatter(
                    x=[ci_mean[0], ci_mean[1], ci_mean[1], ci_mean[0]],
                    y=[idx-0.5, idx-0.5, idx+0.5, idx+0.5],
                    fill='toself',
                    line_width=0,
                    fillcolor=bluebelt.style("equal_means.ci_mean.fill.facecolor"),
                    opacity=bluebelt.style("equal_means.ci_mean.fill.alpha"),
                    mode='lines',
                    showlegend=False if idx > 0 else True,
                    name="CI mean",
                )

            )
            fig.add_trace(
                go.Scatter(
                    x=[mean],
                    y=[idx+0.325],
                    error_x=dict(
                        type='data',
                        symmetric=False,
                        width=0,
                        array=[ci_mean[1]-mean],
                        arrayminus=[mean-ci_mean[0]]),
                    line=dict(color='black'),
                    marker=dict(color='black'),
                    showlegend=False,
                )
            )

            fig.add_annotation(
                x=ci_mean[0],
                y=idx+0.325,
                xref="x",
                yref="y",
                text=f"{ci_mean[0]:1.2f}",
                showarrow=False,
                xanchor="right",
                )
            fig.add_annotation(
                x=ci_mean[1],
                y=idx+0.325,
                xref="x",
                yref="y",
                text=f"{ci_mean[1]:1.2f}",
                showarrow=False,
                xanchor="left",
                )

        # plot popmean if it exists
        if hasattr(plot_obj, "popmean"):
            fig.add_vline(
                x=plot_obj.popmean,
                line_width=bluebelt.style("equal_means.popmean.line.linewidth"),
                line_dash=m2p.linestyle(bluebelt.style("equal_means.popmean.line.linestyle")),
                line_color=bluebelt.style("equal_means.popmean.line.color"),
                )

        fig.update_yaxes(tickvals=list(range(plot_obj.frame.shape[1])), ticktext=plot_obj.frame.columns[::-1])

        return fig
    
    else:
        return

# equal variance
class FTest:
    """
    apply the F-test on a pandas dataframe with the columns as groups to compare
    the F-test is extremely sensitive to non-normality of the two arrays
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """

    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.array_a = self.frame.iloc[:, 0].dropna()
        self.array_b = self.frame.iloc[:, 1].dropna()
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "F-test"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic = self.array_a.var() / self.array_b.var()
        self.p_value = stats.f.cdf(
            self.statistic, self.array_a.size - 1, self.array_b.size - 1
        )
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)


class Levene:
    """
    apply levene's test on a pandas dataframe with the columns as groups to compare
    Returns the scipy LeveneResult; statistic, pvalue
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """

    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "Levene"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.levene(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)


class Bartlett:
    """
    apply Bartlett's test on a pandas dataframe with the columns as groups to compare
    Returns the scipy BartlettResult; statistic, pvalue
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """

    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "Bartlett"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.bartlett(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)


class EqualVariances:
    """
    Test a dataframe for equal variances over the dataframe columns.

    frame: a pandas DataFrame
    columns: string or list of strings with the column names to be tested
    alpha: test threshold

    applies F-test, Bartlett's or Levene's

    pvalue >= alpha (default 0.05) : columns have equal variances
    pvalue < alpha: columns don't have equal variances

    """

    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):

        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        # test for normal distribution; null hypothesis: values come from a normal distribution
        if NormalDistribution(frame=self.frame, alpha=self.alpha).passed:
            # all columns in the dataframe come from a normal distribution
            if self.ncols == 2:
                test = FTest(frame=self.frame, alpha=self.alpha)
            else:
                test = Bartlett(frame=self.frame, alpha=self.alpha)
        else:
            # not all columns in the dataframe come from a normal distribution
            test = Levene(frame=self.frame, alpha=self.alpha)

        self.test = test.test
        self.statistic = test.statistic
        self.p_value = test.p_value
        self.passed = True if self.p_value >= self.alpha else False

    def __str__(self):
        _result = f"-" * (len(self.test) + 12) + "\n"
        _result += f"  {self.test} results\n"
        _result += f"-" * (len(self.test) + 12) + "\n"
        _result += f"\n"

        _result += f"input variables\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"observations:":<30}{self.nrows:1.0f}\n'
        _result += f'  {"alpha:":<30}{self.alpha}\n'

        _result += f"\n"
        _result += f"result\n"
        _result += f"-" * 50 + "\n"
        _result += f'  {"statistic:":<30}{self.statistic:1.4f}\n'
        _result += f'  {"p-value:":<30}{self.p_value:1.4f}\n'
        _result += f'  {"passed:":<30}{self.passed}\n'
        return _result

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)


def _equal_variances_plot(
    plot_obj,
    bins=20,
    xlim=(None, None),
    format_stats=".2f",
    format_xticks=None,
    title=None,
    xlabel=None,
    path=None,
    **kwargs,
):

    title = title if title is not None else f"{plot_obj.test}"
    
    if bluebelt.config('plotting') == 'matplotlib':

        
        # prepare figure
        fig, axes = plt.subplots(
            nrows=plot_obj.ncols,
            ncols=1,
            sharex="col",
            gridspec_kw={"wspace": 0, "hspace": 0},
            **kwargs,
        )

        for row in range(0, plot_obj.ncols):

            if plot_obj.ncols == 1:
                ax = axes
            else:
                ax = axes[row]

            # set right zorder
            ax.set_zorder((plot_obj.ncols - row) * 10)
            # get the data
            array = plot_obj.frame.iloc[:, row]

            # calculate bin width
            bin_width = (
                np.nanmax(plot_obj.frame.values) - np.nanmin(plot_obj.frame.values)
            ) / bins

            # 1. histogram ############################################
            ax.hist(
                array,
                bins=np.arange(
                    np.nanmin(plot_obj.frame.values),
                    np.nanmax(plot_obj.frame.values) + bin_width,
                    bin_width,
                ),
                **bluebelt.style("equal_variances.hist"),
            )
            ax.set_yticks([])
            ax.set_ylabel(array.name, **bluebelt.style("equal_variances.text"))

            if row == 0:
                ax.set_title(
                    f"t={plot_obj.statistic:{format_stats}} p={plot_obj.p_value:{format_stats}}"
                    **bluebelt.style("equal_variances.title"),
                )

        # format things
        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[-1]

        # limit axis
        ax.set_xlim(xlim)

        # format ticks
        if format_xticks:
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])

        # labels
        if title:
            # add suptitle
            fig.suptitle(
                t=title,
                x=fig.subplotpars.left,
                **bluebelt.style("equal_variances.suptitle"),
            )
        if xlabel:
            ax.set_xlabel(xlabel)

        plt.tight_layout()

        # and again
        if title:
            # add suptitle
            fig.suptitle(
                t=title,
                x=fig.subplotpars.left,
                **bluebelt.style("equal_variances.suptitle"),
            )

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    elif bluebelt.config('plotting') == 'plotly':

        fig = make_subplots(
            rows=plot_obj.frame.shape[1],
            cols=1,
            vertical_spacing=0,
            shared_xaxes=True,
            )

        # calculate bins
        bin_data = plot_obj.frame.to_numpy()
        bin_data = bin_data[np.logical_not(np.isnan(bin_data))]
        xbins=dict(start=bin_data.min(), end=bin_data.max(), size=(bin_data.max()-bin_data.min())/bins)

        for idx, col in enumerate(plot_obj.frame.columns):
            fig.add_trace(
                go.Histogram(
                    x=plot_obj.frame[col],
                    histnorm="probability density",
                    xbins=xbins,
                    marker=dict(color=bluebelt.style("equal_variances.hist.facecolor"), opacity=bluebelt.style("equal_variances.hist.alpha")),
                    name=col,
                    opacity=1,
                    showlegend=False,
                ),
                row=idx+1,
                col=1
            )
            fig.update_yaxes(**m2p.yaxis(), row=idx+1, col=1)
            fig.update_yaxes(ticks="", showticklabels=False, row=idx+1, col=1)
            fig.update_xaxes(**m2p.xaxis(), row=idx+1, col=1)
            if idx>0:
                fig.update_xaxes(mirror=False, row=idx+1, col=1)
            
            fig.add_annotation(
                x=0,
                y=0,
                xref="paper",
                yref="paper",
                text=col,
                showarrow=False,
                xanchor="right",
                yanchor="bottom",
                row=idx+1,
                col=1
            )

        if kwargs.get('dash', False):
            title = None
        elif title is not None:
            statistics = f'\\hspace{{1em}} t={plot_obj.statistic:{format_stats}} \\space p={plot_obj.p_value:{format_stats}}'
            title = f'${title} \\hspace{{1em}} {statistics}$'

        fig.update_layout(m2p.layout(title=title, **kwargs))
        fig.update_layout(
            bargap = 1 - bluebelt.style("equal_variances.hist.rwidth"),
            #yaxis = dict(ticks="", showticklabels=False),
        )

        return fig
    
    else:
        return

# distribution
class AndersonDarling:
    """
    Anderson-Darling test

    null hypothesis: x comes from a normal distribution

    test all pandas dataframe columns for normal distribution
    input pivoted dataframe

    pvalue >= alpha : column values are (all) normally distributed
    pvalue < alpha: column values are not (all) normally distributed
    """

    def __init__(self, frame, dist="norm", alpha=0.05, **kwargs):

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)

        # check arguments

        self.frame = frame
        self.dist = dist
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "Anderson-Darling"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        statistics = {}
        p_values = {}
        passed_values = {}

        for col in self.frame.columns:

            AD, critical_values, significance_level = stats.anderson(
                self.frame[col].dropna().values, dist=self.dist
            )

            AD_adjusted = AD * (1 + (0.75 / 50) + 2.25 / (50**2))
            if AD_adjusted >= 0.6:
                pval = math.exp(
                    1.2937 - 5.709 * AD_adjusted - 0.0186 * (AD_adjusted**2)
                )
            elif AD_adjusted >= 0.34:
                pval = math.exp(
                    0.9177 - 4.279 * AD_adjusted - 1.38 * (AD_adjusted**2)
                )
            elif AD_adjusted > 0.2:
                pval = 1 - math.exp(
                    -8.318 + 42.796 * AD_adjusted - 59.938 * (AD_adjusted**2)
                )
            else:
                pval = 1 - math.exp(
                    -13.436 + 101.14 * AD_adjusted - 223.73 * (AD_adjusted**2)
                )

            statistics[col] = AD
            p_values[col] = pval
            passed_values[col] = True if pval >= self.alpha else False

        self.statistics = pd.Series(statistics)
        self.p_values = pd.Series(p_values)
        self.passed_values = pd.Series(passed_values)

        self.statistic = self.statistics[self.p_values.idxmin()]
        self.p_value = self.p_values.min()
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, format_stats=".2f", title=None, path=None, **kwargs):
        return _distribution_plot(
            self, format_stats=format_stats, title=title, path=path, **kwargs
        )


class DAgostinoPearson:
    """
    D’Agostino-Pearson test

    null hypothesis: x comes from a normal distribution

    test all pandas dataframe columns for normal distribution
    input pivoted dataframe
    Returns the scipy NormaltestResult; statistic, pvalue for a single column
    Returns the worst scipy NormaltestResult as a tuple; statistic, pvalue for multiple columns
    pvalue >= alpha : column values are (all) normally distributed
    pvalue < alpha: column values are not (all) normally distributed
    """

    def __init__(self, frame, alpha=0.05, **kwargs):

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)

        self.frame = frame
        self.dist = "norm"
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "D’Agostino-Pearson"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        statistics = {}
        p_values = {}
        passed_values = {}

        statistic = 0
        p_value = 1

        for col in self.frame.columns:
            result = stats.normaltest(self.frame[col].dropna().values)
            if result.pvalue < p_value:
                p_value = result.pvalue
                statistic = result.statistic

            statistics[col] = result.statistic
            p_values[col] = result.pvalue
            passed_values[col] = True if result.pvalue >= self.alpha else False

        self.statistics = pd.Series(statistics)
        self.p_values = pd.Series(p_values)
        self.passed_values = pd.Series(passed_values)

        self.statistic = self.statistics[self.p_values.idxmin()]
        self.p_value = self.p_values.min()
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return f"{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})"

    def plot(self, format_stats=".2f", title=None, path=None, **kwargs):
        return _distribution_plot(
            self, format_stats=format_stats, title=title, path=path, **kwargs
        )


NormalDistribution = DAgostinoPearson


def _distribution_plot(_obj, format_stats=".2f", title=None, path=None, **kwargs):

    title = title if title is not None else  f"{_obj.test} ({_get_distribution(_obj.dist)} distribution)"

    # prepare figure
    ncols = math.ceil(math.sqrt(_obj.ncols))
    nrows = math.ceil(_obj.ncols / ncols)

    fig, axes = plt.subplots(
        nrows=nrows, ncols=ncols, gridspec_kw={"wspace": 0, "hspace": 0}, **kwargs
    )

    for row in range(0, nrows):
        for col in range(0, ncols):
            id = col + (ncols * row)

            if nrows == 1:
                if ncols == 1:
                    ax = axes
                else:
                    ax = axes[col]
            else:
                ax = axes[row, col]

            # get the data
            if id < _obj.ncols:
                array = _obj.frame.iloc[:, id].dropna()
                col = _obj.frame.columns[id]

                # H0 the two distributions are identical, F(x)=G(x)
                parameters = eval("stats." + _obj.dist + ".fit(array.values)")
                (osm, osr), (slope, intercept, r) = stats.probplot(
                    array, dist=eval("stats." + _obj.dist), sparams=parameters, fit=True
                )

                # plot
                ax.plot(osm, osr, **bluebelt.style("distribution_plot.scatter"))
                ax.plot(
                    osm,
                    osm * slope + intercept,
                    **bluebelt.style("distribution_plot.line"),
                )

                # stats
                ax.text(
                    0.98,
                    0.02,
                    f"statistic: {_obj.statistics[col]:{format_stats}}\np-value: {_obj.p_values[col]:{format_stats}}",
                    transform=ax.transAxes,
                    **bluebelt.style("distribution_plot.stats"),
                )

                # plot name
                if _obj.ncols > 1:
                    ax.text(
                        0.02,
                        0.98,
                        f"{_obj.frame.columns[id]}",
                        transform=ax.transAxes,
                        **bluebelt.style("distribution_plot.name"),
                    )

                if id == 0:
                    ax.set_title(title)
                ax.set_xticks([])
                ax.set_yticks([])

                ax.set_ylim(ax.get_xlim())

            else:
                ax.axis("off")

    plt.tight_layout()

    # file
    if path:
        if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


def _data_dict(frame):
    return {k: v.dropna().values for k, v in frame.items()}


def _get_h0_equal_means(_obj, format_stats="1.2f"):
    if len(_obj.frame.columns) == 1:
        result = f"$H_0: \\bar{{X}} = \\mu$"
    elif len(_obj.frame.columns) > 4:
        result = f"$H_0: \\bar{{X}}_{{{_obj.frame.columns[0]}}} = \\bar{{X}}_{{{_obj.frame.columns[1]}}} = \\bar{{X}}_{{...}} = \\bar{{X}}_{{{_obj.frame.columns[-1]}}}$"
    else:
        result = f"$H_0: \\bar{{X}}_{{{_obj.frame.columns[0]}}}$"
        for col in _obj.frame.columns[1:]:
            result += f"$ = \\bar{{X}}_{{{col}}}$"
    
    result += f"   t={_obj.statistic:{format_stats}} p={_obj.p_value:{format_stats}}"

    return result

def _get_h0_equal_means_plotly(_obj, title=None, format_stats="1.2f"):
    if title is None:
        return None
    if len(_obj.frame.columns) == 1:
        result = f"H_0: \\bar{{X}} = \\mu"
    elif len(_obj.frame.columns) > 4:
        result = f"H_0: \\bar{{X}}_{{{_obj.frame.columns[0]}}} = \\bar{{X}}_{{{_obj.frame.columns[1]}}} = \\bar{{X}}_{{...}} = \\bar{{X}}_{{{_obj.frame.columns[-1]}}}"
    else:
        result = f"H_0: \\bar{{X}}_{{{_obj.frame.columns[0]}}}"
        for col in _obj.frame.columns[1:]:
            result += f" = \\bar{{X}}_{{{col}}}"
    
    statistics = f"\\hspace{{1em}} t={_obj.statistic:{format_stats}} \\space p={_obj.p_value:{format_stats}}"
    return f"${title} \\hspace{{1em}} {result}    {statistics}$"



def _get_distribution(dist=None):
    distributions = {
        "norm": "Normal (Gaussian)",
        "alpha": "Alpha",
        "anglit": "Anglit",
        "arcsine": "Arcsine",
        "beta": "Beta",
        "betaprime": "Beta Prime",
        "bradford": "Bradford",
        "burr": "Burr",
        "cauchy": "Cauchy",
        "chi": "Chi",
        "chi2": "Chi-squared",
        "cosine": "Cosine",
        "dgamma": "Double Gamma",
        "dweibull": "Double Weibull",
        "erlang": "Erlang",
        "expon": "Exponential",
        "exponweib": "Exponentiated Weibull",
        "exponpow": "Exponential Power",
        "fatiguelife": "Fatigue Life (Birnbaum-Sanders)",
        "foldcauchy": "Folded Cauchy",
        "f": "F (Snecdor F)",
        "fisk": "Fisk",
        "foldnorm": "Folded Normal",
        "frechet_r": "Frechet Right Sided, Extreme Value Type II",
        "frechet_l": "Frechet Left Sided, Weibull_max",
        "gamma": "Gamma",
        "gausshyper": "Gauss Hypergeometric",
        "genexpon": "Generalized Exponential",
        "genextreme": "Generalized Extreme Value",
        "gengamma": "Generalized gamma",
        "genlogistic": "Generalized Logistic",
        "genpareto": "Generalized Pareto",
        "genhalflogist": "Generalized Half Logistic",
        "gilbrat": "Gilbrat",
        "gompertz": "Gompertz (Truncated Gumbel)",
        "gumbel_l": "Left Sided Gumbel, etc.",
        "gumbel_r": "Right Sided Gumbel",
        "halfcauchy": "Half Cauchy",
        "halflogistic": "Half Logistic",
        "halfnorm": "Half Normal",
        "hypsecant": "Hyperbolic Secant",
        "invgamma": "Inverse Gamma",
        "invnorm": "Inverse Normal",
        "invweibull": "Inverse Weibull",
        "johnsonsb": "Johnson SB",
        "johnsonsu": "Johnson SU",
        "laplace": "Laplace",
        "logistic": "Logistic",
        "loggamma": "Log-Gamma",
        "loglaplace": "Log-Laplace (Log Double Exponential",
        "lognorm": "Log-Normal",
        "lomax": "Lomax (Pareto of the second kind)",
        "maxwell": "Maxwell",
        "mielke": "Mielke's Beta-Kappa",
        "nakagami": "Nakagami",
        "ncx2": "Non-central chi-squared",
        "ncf": "Non-central F",
        "nct": "Non-central Student's T",
        "pareto": "Pareto",
        "powerlaw": "Power-function",
        "powerlognorm": "Power log normal",
        "powernorm": "Power normal",
        "rdist": "R distribution",
        "reciprocal": "Reciprocal",
        "rayleigh": "Rayleigh",
        "rice": "Rice",
        "recipinvgauss": "Reciprocal Inverse Gaussian",
        "semicircular": "Semicircular",
        "t": "Student's T",
        "triang": "Triangular",
        "truncexpon": "Truncated Exponential",
        "truncnorm": "Truncated Normal",
        "tukeylambda": "Tukey-Lambda",
        "uniform": "Uniform",
        "vonmises": "Von-Mises (Circular)",
        "wald": "Wald",
        "weibull_min": "Minimum Weibull (see Frechet)",
        "weibull_max": "Maximum Weibull (see Frechet)",
        "wrapcauchy": "Wrapped Cauchy",
        "ksone": "Kolmogorov-Smirnov one-sided (no st",
        "kstwobign": "Kolmogorov-Smirnov two-sided test for Large N",
    }

    return distributions.get(dist, None)



import numpy as np
import pandas as pd

def _get_boxplot_quantiles(series, whis=1.5):
    # get the matplotlib boxplot quantiles
    # sort of direct copy from the matplotlib library

    q1, median, q3 = np.percentile(series, [25, 50, 75])

    if np.isscalar(whis):
        iqr = q3 - q1
        loval = q1 - (whis * iqr)
        hival = q3 + (whis * iqr)
    else:
        loval, hival = np.percentile(series, sorted(whis))

    # get high extreme
    wiskhi = series[series <= hival]
    if len(wiskhi) == 0 or np.max(wiskhi) < q3:
        high = q3
    else:
        high = np.max(wiskhi)

    # get low extreme
    wisklo = series[series >= loval]
    if len(wisklo) == 0 or np.min(wisklo) > q1:
        low = q1
    else:
        low = np.min(wisklo)
    
    return [low, q1, q3, high]

def _get_boxplot_outliers(series, whis=1.5):
    low, q1, q3, high = _get_boxplot_quantiles(series=series, whis=whis)

    return pd.concat([series[series < low],series[series > high]]).values



def _get_nice_list(the_list, max_len=3):
    the_list = [str(item) for item in the_list] # just in case they are not strings
    if len(the_list) == 1:
        return the_list[0]
    elif len(the_list) <= max_len:
        return ", ".join(the_list[:-1])+" and "+the_list[-1]
    else:
        return ", ".join(the_list[:max_len-1])+", ... and "+the_list[-1]


import os
import bluebelt
import pandas as pd

import bluebelt.helpers.ticks
import bluebelt.data.resolution
import matplotlib.pyplot as plt

import plotly.graph_objects as go
import bluebelt.helpers.matplotlib2plotly as m2p

from bluebelt.helpers.decorators.performance import performance
from bluebelt.helpers.ci import ci_mean

# TO DO
# -----
# add line plot for volume % per weekday per week
# add line plot for skill volume % per week and/or date

@performance
class Effort():
    def __init__(self, _obj, shift=1, skip_na=True, skip_zero=True):  
        self._obj = _obj
        self.quantity = bluebelt.data.resolution.Resample(self._obj, 7).diff_quantity(shift=shift, skip_na=skip_na, skip_zero=skip_zero)
        self.distribution = bluebelt.data.resolution.Resample(self._obj, 7).diff_distribution(shift=shift, skip_na=skip_na, skip_zero=skip_zero)
        self.skills = bluebelt.data.resolution.Resample(self._obj, 7).diff_skills(shift=shift, skip_na=skip_na, skip_zero=skip_zero)
        self.qds = 1 - ((1 - self.quantity) * (1 - self.distribution) * (1 - self.skills))

    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self._obj.shape[0]:1.0f}, qds={self.qds.mean():1.4f}, quantity={self.quantity.mean():1.4f}, distribution={self.distribution.mean():1.4f}, skills={self.skills.mean():1.4f})')
    
    def plot(self, xlim=(None, None), ylim=(0, 1), max_xticks=None, format_xticks=None, format_yticks=".0%", format_stats=".1%", title=None, xlabel=None, ylabel=None, legend=True, path=None, **kwargs):
        
        title = title if title is not None else  f'planning effort'

        if bluebelt.config('plotting') == 'matplotlib':

            # prepare figure
            fig, ax = plt.subplots(nrows=1, ncols=1, **kwargs)

            # # q
            ax.fill_between(self.quantity.index[1:], 0, self.quantity.values[1:], label=f'quantity ({self.quantity.mean():{format_stats}})', **bluebelt.style('effort.quantity.fill'))
            ax.plot(self.quantity.index[1:], self.quantity.values[1:], **bluebelt.style('effort.quantity.line'))
            
            # d
            ax.fill_between(self.distribution.index[1:], 0, self.distribution.values[1:], label=f'distribution ({self.distribution.mean():{format_stats}})', **bluebelt.style('effort.distribution.fill'))
            ax.plot(self.distribution.index[1:], self.distribution.values[1:], **bluebelt.style('effort.distribution.line'))

            # s
            ax.fill_between(self.skills.index[1:], 0, self.skills.values[1:], label=f'skills ({self.skills.mean():{format_stats}})', **bluebelt.style('effort.skills.fill'))
            ax.plot(self.skills.index[1:], self.skills.values[1:], **bluebelt.style('effort.skills.line'))
            
            # qds
            ax.plot(self.qds.index[1:], self.qds.values[1:], label=f'qds effort ({self.qds.mean():{format_stats}})', **bluebelt.style('effort.qds.line'))
            
            # limit axis
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            
            # set xticks
            if max_xticks is None:
                max_xticks = bluebelt.helpers.ticks.get_max_xticks(ax)
            bluebelt.helpers.ticks.year_week(self.qds, ax=ax, max_xticks=max_xticks)

            # format ticks
            if format_xticks:
                ax.set_xticks(ax.get_xticks())
                ax.set_xticklabels([f'{x:{format_xticks}}' for x in ax.get_xticks()])
            if format_yticks:
                ax.set_yticks(ax.get_yticks())
                ax.set_yticklabels([f'{y:{format_yticks}}' for y in ax.get_yticks()])

            # labels
            if title:
                ax.set_title(title)
            if xlabel:
                ax.set_xlabel(xlabel)
            if ylabel:
                ax.set_ylabel(ylabel)

            # legend
            if legend:
                ax.legend()
            elif ax.get_legend() is not None:
                ax.get_legend().set_visible(False)
            
            plt.tight_layout()

            # file
            if path:
                if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))
                plt.savefig(path)
                plt.close()
            else:
                plt.close()
                return fig

        elif bluebelt.config('plotting') == 'plotly':
            
            layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    ylim=ylim,
                    format_xticks=format_xticks,
                    format_yticks=format_yticks,
                    legend=legend,
                    **kwargs,
                )
            )
            
            fig = go.Figure(layout=layout)
            
            # quantity
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.quantity.index[1:],
                    y=self.quantity.values[1:],
                    name=f'quantity ({self.quantity.mean():{format_stats}})',
                    line = dict(color=bluebelt.style('effort.quantity.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    fill='tozeroy',
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            # distribution
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.distribution.index[1:],
                    y=self.distribution.values[1:],
                    name=f'distribution ({self.distribution.mean():{format_stats}})',
                    line = dict(color=bluebelt.style('effort.distribution.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    fill='tozeroy',
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            # skills
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.skills.index[1:],
                    y=self.skills.values[1:],
                    name=f'skills ({self.skills.mean():{format_stats}})',
                    line = dict(color=bluebelt.style('effort.skills.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    fill='tozeroy',
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            # qds
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.qds.index[1:],
                    y=self.qds.values[1:],
                    name=f'qds ({self.qds.mean():{format_stats}})',
                    line = dict(color=bluebelt.style('effort.qds.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            return fig
        
class Development():

    def __init__(self, _obj, min_count=2, skip_na=True, skip_zero=True):
        self._obj = _obj
        
        quantity = pd.DataFrame(dtype=float)
        distribution = pd.DataFrame(dtype=float)
        skills = pd.DataFrame(dtype=float)
        qds = pd.DataFrame(dtype=float)

        for name, shift in enumerate(range(1, len(self._obj.index.isocalendar().groupby(['year', 'week']).groups)-min_count)):
            effort = self._obj._.planning.effort(shift=shift)
            _quantity = {
                'lower': ci_mean(effort.quantity)[0],
                'mean': effort.quantity.mean(),
                'upper': ci_mean(effort.quantity)[1],
            }
            _distribution = {
                'lower': ci_mean(effort.distribution)[0],
                'mean': effort.distribution.mean(),
                'upper': ci_mean(effort.distribution)[1],
            }
            _skills = {
                'lower': ci_mean(effort.skills)[0],
                'mean': effort.skills.mean(),
                'upper': ci_mean(effort.skills)[1],
            }
            _qds = {
                'lower': ci_mean(effort.qds)[0],
                'mean': effort.qds.mean(),
                'upper': ci_mean(effort.qds)[1],
            }
            quantity = pd.concat([quantity, pd.Series(_quantity, name=name+1)], axis=1)
            distribution = pd.concat([distribution, pd.Series(_distribution, name=name+1)], axis=1)
            skills = pd.concat([skills, pd.Series(_skills, name=name+1)], axis=1)
            qds = pd.concat([qds, pd.Series(_qds, name=name+1)], axis=1)
            
        self.quantity = quantity.T
        self.distribution = distribution.T
        self.skills = skills.T
        self.qds = qds.T

    def plot(self, xlim=(None, None), ylim=(0, 1), max_xticks=None, format_xticks=None, format_yticks=".0%", title=None, xlabel=None, ylabel=None, legend=True, path=None, **kwargs):

        title = title if title is not None else  f'planning effort development'
        
        if bluebelt.config('plotting') == 'matplotlib':
            # prepare figure
            fig, ax = plt.subplots(nrows=1, ncols=1, **kwargs)

            # q
            ax.fill_between(self.quantity.index, self.quantity['lower'], self.quantity['upper'], label=f'quantity', **bluebelt.style('effort.quantity.fill'))
            ax.plot(self.quantity.index, self.quantity['mean'], **bluebelt.style('effort.quantity.line'))
            
            # d
            ax.fill_between(self.distribution.index, self.distribution['lower'], self.distribution['upper'], label=f'distribution', **bluebelt.style('effort.distribution.fill'))
            ax.plot(self.distribution.index, self.distribution['mean'], **bluebelt.style('effort.distribution.line'))
            
            # s
            ax.fill_between(self.skills.index, self.skills['lower'], self.skills['upper'], label=f'skills', **bluebelt.style('effort.skills.fill'))
            ax.plot(self.skills.index, self.skills['mean'], **bluebelt.style('effort.skills.line'))
            
            # qds
            ax.fill_between(self.qds.index, self.qds['lower'], self.qds['upper'], label=f'qds', **bluebelt.style('effort.qds.fill'))
            ax.plot(self.qds.index, self.qds['mean'], **bluebelt.style('effort.qds.line'))
            
            # limit axis
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            
            # set xticks
            if max_xticks is None:
                max_xticks = bluebelt.helpers.ticks.get_max_xticks(ax)
            #bluebelt.helpers.ticks.year_week(self.qds, ax=ax, max_xticks=max_xticks)

            # format ticks
            if format_xticks:
                ax.set_xticks(ax.get_xticks())
                ax.set_xticklabels([f'{x:{format_xticks}}' for x in ax.get_xticks()])
            if format_yticks:
                ax.set_yticks(ax.get_yticks())
                ax.set_yticklabels([f'{y:{format_yticks}}' for y in ax.get_yticks()])

            # labels
            if title:
                ax.set_title(title)
            if xlabel:
                ax.set_xlabel(xlabel)
            if ylabel:
                ax.set_ylabel(ylabel)

            # legend
            if legend:
                ax.legend()
            elif ax.get_legend() is not None:
                ax.get_legend().set_visible(False)
            
            plt.tight_layout()

            # file
            if path:
                if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                    os.makedirs(os.path.dirname(path))
                plt.savefig(path)
                plt.close()
            else:
                plt.close()
                return fig

        elif bluebelt.config('plotting') == 'plotly':

            layout = go.Layout(
                    m2p.layout(
                        title=title,
                        xlabel=xlabel,
                        ylabel=ylabel,
                        xlim=xlim,
                        ylim=ylim,
                        format_xticks=format_xticks,
                        format_yticks=format_yticks,
                        legend=legend,
                        **kwargs,
                    )
                )

            fig = go.Figure(layout=layout) 

            
            # qds
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.qds.index,
                    y=self.qds['lower'],
                    line = dict(color=bluebelt.style('effort.qds.line.color'), width=bluebelt.style('effort.qds.line.linewidth')),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.qds.index,
                    y=self.qds['upper'],
                    line = dict(color=bluebelt.style('effort.qds.line.color'), width=bluebelt.style('effort.qds.line.linewidth')),
                    fill='tonexty',
                    fillcolor=m2p.to_rgba(
                        bluebelt.style('effort.qds.fill.color'),
                        bluebelt.style('effort.qds.fill.alpha')
                        ),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.qds.index,
                    y=self.qds['mean'],
                    name=f'qds',
                    line = dict(color=bluebelt.style('effort.qds.line.color'), width=bluebelt.style('effort.qds.line.linewidth')),
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            # skills
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.skills.index,
                    y=self.skills['lower'],
                    line = dict(color=bluebelt.style('effort.skills.line.color'), width=bluebelt.style('effort.skills.line.linewidth')),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.skills.index,
                    y=self.skills['upper'],
                    line = dict(color=bluebelt.style('effort.skills.line.color'), width=bluebelt.style('effort.skills.line.linewidth')),
                    fill='tonexty',
                    fillcolor=m2p.to_rgba(
                        bluebelt.style('effort.skills.fill.color'),
                        bluebelt.style('effort.skills.fill.alpha')
                        ),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.skills.index,
                    y=self.skills['mean'],
                    name=f'skills',
                    line = dict(color=bluebelt.style('effort.skills.line.color'), width=bluebelt.style('effort.skills.line.linewidth')),
                    hoverinfo="none",
                    showlegend=True,
                ),
            )
            
            # distribution
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.distribution.index,
                    y=self.distribution['lower'],
                    line = dict(color=bluebelt.style('effort.distribution.line.color'), width=bluebelt.style('effort.distribution.line.linewidth')),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.distribution.index,
                    y=self.distribution['upper'],
                    line = dict(color=bluebelt.style('effort.distribution.line.color'), width=bluebelt.style('effort.distribution.line.linewidth')),
                    fill='tonexty',
                    fillcolor=m2p.to_rgba(
                        bluebelt.style('effort.distribution.fill.color'),
                        bluebelt.style('effort.distribution.fill.alpha')
                        ),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.distribution.index,
                    y=self.distribution['mean'],
                    name=f'distribution',
                    line = dict(color=bluebelt.style('effort.distribution.line.color'), width=bluebelt.style('effort.distribution.line.linewidth')),
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            # quantity
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.quantity.index,
                    y=self.quantity['lower'],
                    line = dict(color=bluebelt.style('effort.quantity.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.quantity.index,
                    y=self.quantity['upper'],
                    line = dict(color=bluebelt.style('effort.quantity.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    fill='tonexty',
                    fillcolor=m2p.to_rgba(
                        bluebelt.style('effort.quantity.fill.color'),
                        bluebelt.style('effort.quantity.fill.alpha')
                        ),
                    hoverinfo="none",
                    showlegend=False,
                ),
            )
            fig.add_trace(
                go.Scatter(
                    mode="lines",
                    x=self.quantity.index,
                    y=self.quantity['mean'],
                    name=f'quantity',
                    line = dict(color=bluebelt.style('effort.quantity.line.color'), width=bluebelt.style('effort.quantity.line.linewidth')),
                    hoverinfo="none",
                    showlegend=True,
                ),
            )

            return fig
        
        else:
            return        



import numpy as np
import pandas as pd

import bluebelt.helpers.date

import datetime

def project(_obj, year=None, adjust_holidays=True):
    """
        Project an object with DatetimeIndex to another year. If the object
        contains more than one iso year the data will be projected from the
        provided year up.
        Data is projected by iso week and weekday; e.g. the data at Monday
        of isoweek 1 in the original year data is projected to Monday of
        iso week 1 in the projection year.
    
        Parameters
        ----------
        year: int, default None
            the year to project to
        adjust_holidays: boolean, default True
            adjust the values for holidays; take the values for each holiday
            and swap them with the new holiday dates

        Returns
        -------
        The projected pandas object

    """
    _obj = _obj.copy()

    # create result object
    if isinstance(_obj, pd.Series):
        result = pd.Series(dtype=_obj.dtype)
    else:
        result = pd.DataFrame(dtype=_obj.dtype)

    # check if _obj has only one iso year
    for this_year in _obj.index.isocalendar().year.unique():

        year_obj = _obj[_obj.index.isocalendar().year == this_year]

        _year = year + this_year - _obj.index.isocalendar().year.min()

        # check for week 53
        if (
            year_obj.index.isocalendar().week.max()
            > bluebelt.helpers.date.last_iso_week_in_year(_year)
        ):
            # drop week 53
            year_obj = year_obj[year_obj.index.isocalendar().week < 53]
        elif (
            year_obj.index.isocalendar().week.max()
            < bluebelt.helpers.date.last_iso_week_in_year(_year)
        ):
            # TO DO OPTIONAL create week 53
            pass

        # create a isoformat MultiIndex
        index = pd.MultiIndex.from_frame(year_obj.index.isocalendar())
        # set the year to the requested year
        index = index.set_levels([_year], level=0)
        # build a new DatetimeIndex from the isoformat MultiIndex
        index = pd.DatetimeIndex(
            np.apply_along_axis(
                lambda x: datetime.datetime.fromisocalendar(x[0], x[1], x[2]),
                1,
                np.array([*index.values]),
            )
        )
        # create the new year_obj
        if isinstance(year_obj, pd.Series):
            result = pd.concat([result, pd.Series(index=index, data=year_obj.values, name=year_obj.name)])
        else:
            result = pd.concat([result, pd.DataFrame(index=index, data=year_obj.values, columns=year_obj.columns)])

        if adjust_holidays:

            # get original holidays
            _obj_holidays = bluebelt.helpers.holidays.get_holidays_dict(
                year_obj.index.year.unique()
            )
            # find the new dates for the original holidays
            _obj_holidays = {
                datetime.datetime.fromisocalendar(
                    _year, key.isocalendar()[1], key.isocalendar()[2]
                ): value
                for key, value in _obj_holidays.items()
                if key in year_obj.index
            }

            # get new holidays
            result_holidays = bluebelt.helpers.holidays.get_holidays_dict(
                result.index.year.unique()
            )
            result_holidays = {
                value: key for key, value in result_holidays.items() if key in result.index
            }

            # create a swap dictionairy
            holidays = {
                key: result_holidays.get(value) for key, value in _obj_holidays.items()
            }

            for key, value in holidays.items():

                # catch none existing holidays like liberation day in some years
                if key and value:
                    result.loc[[key, value]] = result.loc[[value, key]].values

    return result



import os
import pandas as pd
import numpy as np
import scipy.stats as stats

import matplotlib.pyplot as plt
import plotly.graph_objects as go
import bluebelt.helpers.matplotlib2plotly as m2p

import bluebelt.statistics.hypothesis_testing

import bluebelt.styles


import bluebelt.helpers.language


class WeekDay:
    def __init__(self, series, threshold=8, **kwargs):

        self.series = series
        self.threshold = threshold
        self.name = series.name
        self.calculate()

    def calculate(self):
        self.series = pd.Series(
            index=self.series.index.weekday,
            data=self.series.values,
            name=self.series.name,
        ).sort_index(
            level=0
        )  # .droplevel(0)

        _calculate(self)

        columns = bluebelt.helpers.language.weekdays.get(bluebelt.config("language"))

        self.frame = self.frame.rename(columns=columns)

    def __repr__(self):
        return f"{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})"

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title="week day distribution",
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )


class MonthDay:
    """
    Compare the distribution of data between month days
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex

        properties
        .series
            the transformed pandas.Series with month day index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """

    def __init__(self, series, threshold=8, **kwargs):

        self.series = series
        self.threshold = threshold
        self.name = series.name
        self.calculate()

    def calculate(self):

        self.series = pd.Series(
            index=self.series.index.day, data=self.series.values, name=self.series.name
        ).sort_index()

        _calculate(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})"

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title="month day distribution",
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )


class Week:
    """
    Compare the distribution of data between weeks
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex

        properties
        .series
            the transformed pandas.Series with week number index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """

    def __init__(self, series, threshold=8, **kwargs):

        self.series = series
        self.threshold = threshold
        self.name = series.name
        self.calculate()

    def calculate(self):

        self.series = pd.Series(
            index=self.series.index.isocalendar().week,
            data=self.series.values,
            name=self.series.name,
        ).sort_index()

        _calculate(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})"

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title="week distribution",
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )


class Month:
    """
    Compare the distribution of data between months
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex

        properties
        .series
            the transformed pandas.Series with month index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """

    def __init__(self, series, threshold=8, **kwargs):

        self.series = series
        self.threshold = threshold
        self.name = series.name
        self.calculate()

    def calculate(self):

        self.series = pd.Series(
            index=self.series.index.month,
            data=self.series.values,
            name=self.series.name,
        ).sort_index()

        _calculate(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})"

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title="month distribution",
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )


class Year:
    """
    Compare the distribution of data between years
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex

        properties
        .series
            the transformed pandas.Series with year index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """

    def __init__(self, series, threshold=8, **kwargs):

        self.series = series
        self.threshold = threshold
        self.name = series.name
        self.calculate()

    def calculate(self):

        self.series = pd.Series(
            index=self.series.index.year, data=self.series.values, name=self.series.name
        ).sort_index()

        _calculate(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})"

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title="year distribution",
        xlabel=None,
        ylabel=None,
        path=None,
        **kwargs,
    ):
        return _plot(
            self,
            xlim=xlim,
            ylim=ylim,
            max_xticks=max_xticks,
            format_xticks=format_xticks,
            format_yticks=format_yticks,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            path=path,
            **kwargs,
        )


def _calculate(dt_obj, **kwargs):
    # Why orient='index' and then .T you ask? Simple; the lists are not of equal length and this now gets solved.
    dt_obj.frame = pd.DataFrame.from_dict(
        {
            i: pd.Series(dt_obj.series[i]).to_list()
            for i in dt_obj.series.index.unique()
            if dt_obj.series[dt_obj.series.index == i].count() >= dt_obj.threshold
        },
        orient="index",
    ).T
    dt_obj.means = dt_obj.frame.mean()
    dt_obj.ratio = dt_obj.means / dt_obj.means.mean()
    dt_obj.variances = dt_obj.frame.var()
    dt_obj.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(
        dt_obj.frame
    ).passed
    dt_obj.equal_variances = bluebelt.statistics.hypothesis_testing.EqualVariances(
        dt_obj.frame
    ).passed

    normal_distribution_test = (
        bluebelt.statistics.hypothesis_testing.NormalDistribution(dt_obj.frame)
    )

    dt_obj.normal_distribution = normal_distribution_test.passed_values
    dt_obj.normal_distribution_p_values = normal_distribution_test.p_values


def _plot(
    plot_obj,
    xlim=(None, None),
    ylim=(None, None),
    max_xticks=None,
    format_xticks=None,
    format_yticks=None,
    title=None,
    xlabel=None,
    ylabel=None,
    legend=False,
    path=None,
    **kwargs,
):

    title = title if title is not None else  f"{plot_obj.name} datetime distribution"

    if bluebelt.config('plotting') == 'matplotlib':

        # prepare figure
        fig, ax = plt.subplots(**kwargs)

        boxplot = ax.boxplot([plot_obj.frame[col].dropna() for col in plot_obj.frame])

        # boxes
        for n, box in enumerate(boxplot["boxes"]):
            # add style if any is given
            box.set(**bluebelt.style("datetime.boxplot.boxes"))

        # whiskers
        for n, whisker in enumerate(boxplot["whiskers"]):
            # add style if any is given
            whisker.set(**bluebelt.style("datetime.boxplot.whiskers"))

        # caps
        for n, cap in enumerate(boxplot["caps"]):
            # add style if any is given
            whisker.set(**bluebelt.style("datetime.boxplot.caps"))

        # limit axis
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

        # set ticks
        if max_xticks is None:
            max_xticks = bluebelt.helpers.ticks.get_max_xticks(ax)
        bluebelt.helpers.ticks.ticks(
            plot_obj.frame.T, ax=ax, max_xticks=max_xticks
        )  # we need the column titles as ticks

        # format ticks
        if format_xticks:
            ax.set_xticks(ax.get_xticks())
            ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])
        if format_yticks:
            ax.set_yticks(ax.get_yticks())
            ax.set_yticklabels([f"{y:{format_yticks}}" for y in ax.get_yticks()])

        # labels
        if title:
            ax.set_title(title)
        if xlabel:
            ax.set_xlabel(xlabel)
        if ylabel:
            ax.set_ylabel(ylabel)

        plt.tight_layout()

        # file
        if path:
            if len(os.path.dirname(path)) > 0 and not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig
    elif bluebelt.config('plotting') == 'plotly':
        points = kwargs.get('points', False)

        layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    ylim=ylim,
                    format_xticks=format_xticks,
                    format_yticks=format_yticks,
                    legend=legend,
                    **kwargs,
                )
            )

        line_width = np.maximum(1, bluebelt.style("datetime.boxplot.boxes.linewidth"))
        data = [
          
            go.Box(
                y=plot_obj.frame[col].values,
                marker=dict(color=bluebelt.style("datetime.boxplot.fliers.color"), symbol=m2p.marker(bluebelt.style("datetime.boxplot.fliers.marker"))),
                fillcolor=bluebelt.style("datetime.boxplot.boxes.facecolor"),
                line_color=bluebelt.style("datetime.boxplot.boxes.edgecolor"),
                line_width=line_width,
                hoverinfo='skip',
                name=col,
                
            ) for col in plot_obj.frame.columns
        ]

        fig = go.Figure(data=data, layout=layout)

        if points:
            fig.update_traces(
                jitter=0.3,
                pointpos=-1.8,
                boxpoints='all',
            )        

        return fig
    
    else:
        return


import numpy as np
import scipy.stats as stats

import warnings

def get_distributions(sample):
    cdfs = [
    "norm",            #Normal (Gaussian)
    "alpha",           #Alpha
    "anglit",          #Anglit
    "arcsine",         #Arcsine
    "beta",            #Beta
    "betaprime",       #Beta Prime
    "bradford",        #Bradford
    "burr",            #Burr
    "cauchy",          #Cauchy
    "chi",             #Chi
    "chi2",            #Chi-squared
    "cosine",          #Cosine
    "dgamma",          #Double Gamma
    "dweibull",        #Double Weibull
    "erlang",          #Erlang
    "expon",           #Exponential
    "exponweib",       #Exponentiated Weibull
    "exponpow",        #Exponential Power
    "fatiguelife",     #Fatigue Life (Birnbaum-Sanders)
    "foldcauchy",      #Folded Cauchy
    "f",               #F (Snecdor F)
    "fisk",            #Fisk
    "foldnorm",        #Folded Normal
    "frechet_r",       #Frechet Right Sided, Extreme Value Type II
    "frechet_l",       #Frechet Left Sided, Weibull_max
    "gamma",           #Gamma
    "gausshyper",      #Gauss Hypergeometric
    "genexpon",        #Generalized Exponential
    "genextreme",      #Generalized Extreme Value
    "gengamma",        #Generalized gamma
    "genlogistic",     #Generalized Logistic
    "genpareto",       #Generalized Pareto
    "genhalflogistic", #Generalized Half Logistic
    "gilbrat",         #Gilbrat
    "gompertz",        #Gompertz (Truncated Gumbel)
    "gumbel_l",        #Left Sided Gumbel, etc.
    "gumbel_r",        #Right Sided Gumbel
    "halfcauchy",      #Half Cauchy
    "halflogistic",    #Half Logistic
    "halfnorm",        #Half Normal
    "hypsecant",       #Hyperbolic Secant
    "invgamma",        #Inverse Gamma
#    "invnorm",         #Inverse Normal
    "invweibull",      #Inverse Weibull
    "johnsonsb",       #Johnson SB
    "johnsonsu",       #Johnson SU
    "laplace",         #Laplace
    "logistic",        #Logistic
    "loggamma",        #Log-Gamma
    "loglaplace",      #Log-Laplace (Log Double Exponential)
    "lognorm",         #Log-Normal
    "lomax",           #Lomax (Pareto of the second kind)
    "maxwell",         #Maxwell
    "mielke",          #Mielke's Beta-Kappa
    "nakagami",        #Nakagami
    "ncx2",            #Non-central chi-squared
#    "ncf",             #Non-central F
    "nct",             #Non-central Student's T
    "pareto",          #Pareto
    "powerlaw",        #Power-function
    "powerlognorm",    #Power log normal
    "powernorm",       #Power normal
    "rdist",           #R distribution
    "reciprocal",      #Reciprocal
    "rayleigh",        #Rayleigh
    "rice",            #Rice
    "recipinvgauss",   #Reciprocal Inverse Gaussian
    "semicircular",    #Semicircular
    "t",               #Student's T
    "triang",          #Triangular
    "truncexpon",      #Truncated Exponential
    "truncnorm",       #Truncated Normal
    "tukeylambda",     #Tukey-Lambda
    "uniform",         #Uniform
    "vonmises",        #Von-Mises (Circular)
    "wald",            #Wald
    "weibull_min",     #Minimum Weibull (see Frechet)
    "weibull_max",     #Maximum Weibull (see Frechet)
    "wrapcauchy",      #Wrapped Cauchy
    "ksone",           #Kolmogorov-Smirnov one-sided (no stats)
    "kstwobign"]       #Kolmogorov-Smirnov two-sided test for Large N
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for cdf in cdfs:
            #fit our data set against every probability distribution
            parameters = eval("stats."+cdf+".fit(sample)");

            #Applying the Kolmogorov-Smirnof one sided test
            D, p = stats.kstest(sample, cdf, args=parameters);

            #pretty-print the results
            print (cdf.ljust(16) + ("p: "+str(p)).ljust(25)+"D: "+str(D))